#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import gi

gi.require_version('Gtk', '3.0')

from gi.repository import Gio, Gtk, Gdk, GObject, GdkPixbuf
from bs4 import BeautifulSoup
import types
import psutil
import os
import sys
import apt
import struct
import gettext
import pipes
import re
import aptdaemon.client
import aptdaemon.errors
import aptdaemon.gtk3widgets
import xml.etree.ElementTree
from .appdata import *

def moduleVar(strID):

    return 'm_p' + strID[0:1].upper() + strID[1:]

def quote(str):

    if len(str) == 0:
        return "''"
    else:
        return pipes.quote(str)

def get_data_path(strPath):

    try:

        strExecPath = os.path.split(APPEXECUTABLE)[0]
        strDataPath = os.getcwd().replace(strExecPath, '')
        strRelativePath = os.path.join(strDataPath, strPath.lstrip('/'))

        if os.path.exists(strRelativePath):
            return strRelativePath

    except:

        pass

    return strPath

def notification(oNotify, strSummary, strBody):

    oNotification = oNotify.Notification.new(strSummary, strBody, None)
    oNotification.set_image_from_pixbuf(GdkPixbuf.Pixbuf().new_from_file(get_data_path('/usr/share/icons/hicolor/scalable/apps/' + APPNAME + '.svg')))
    oNotification.show()

try:
    g_oTranslation = gettext.translation(APPNAME)
except IOError:
    g_oTranslation = gettext.NullTranslations()

g_oTranslation.install()

g_pSettings = Gio.Settings.new('in.tari.' + APPNAME)
g_pWindow = None

class SplashScreen(Gtk.Window):

    def __init__(self, fnName, bShowSplashScreen):

        self.bShowSplashScreen = bShowSplashScreen
        self.fnName = fnName

        if self.bShowSplashScreen:

            Gtk.Window.__init__(self)
            self.set_title(APPTITLE)
            self.set_skip_taskbar_hint(True)
            self.set_skip_pager_hint(True)
            self.set_modal(True)
            self.set_decorated(False)
            self.set_type_hint(Gdk.WindowTypeHint.SPLASHSCREEN)
            self.set_position(Gtk.WindowPosition.CENTER)
            self.set_size_request(480, 148)
            self.set_resizable(False)
            hbox = Gtk.HBox(False, 10)
            hbox.set_border_width(10)
            self.add(hbox)
            image = Gtk.Image()
            image.set_from_file(get_data_path('/usr/share/icons/hicolor/scalable/apps/' + APPNAME + '.svg'))
            hbox.pack_start(image, False, True, 0)
            self.label = Gtk.Label(_('Resolving dependencies...'))
            self.label.set_justify(Gtk.Justification.CENTER)
            hbox.pack_start(self.label, True, True, 1)

            self.realize()
            self.set_transient_for(None)
            self.grab_focus()
            self.show_all()

            while Gtk.events_pending():
                Gtk.main_iteration()

        arrDependenciesLocal = []
        arrDependenciesExternal = []

        if APPDEPENDENCIES:

            pCache = apt.Cache()

            for strDependency in APPDEPENDENCIES:

                if strDependency.endswith(' [i386]'):
                    if 8 * struct.calcsize('P') == 32:
                        strDependency = strDependency.replace(' [i386]', '')
                    else:
                        continue
                elif strDependency.endswith(' [amd64]'):
                    if 8 * struct.calcsize('P') == 64:
                        strDependency = strDependency.replace(' [amd64]', '')
                    else:
                        continue

                if strDependency in pCache:

                    if pCache[strDependency].is_installed == False:
                        arrDependenciesLocal.append(strDependency)

                elif os.path.isfile('/usr/bin/' + strDependency) == False:
                    arrDependenciesExternal.append(strDependency)

            if len(arrDependenciesLocal):

                dlg = Gtk.MessageDialog(None, Gtk.DialogFlags.MODAL, Gtk.MessageType.INFO, Gtk.ButtonsType.OK, _('The following dependencies must be installed in order to use this application:') + '\n\n' + '\n'.join(arrDependenciesLocal))
                dlg.run()
                dlg.destroy()

                pAptClient = aptdaemon.client.AptClient()
                self.pTransaction = pAptClient.install_packages(arrDependenciesLocal, error_handler=self.on_error)
                self.pTransaction.connect('finished', self.on_transaction_finished)
                pAptProgressDialog = aptdaemon.gtk3widgets.AptProgressDialog(self.pTransaction, parent=self)
                pAptProgressDialog.run(close_on_finished=True, show_error=True, reply_handler=lambda: True, error_handler=self.on_error)

            if len(arrDependenciesExternal):

                dlg = Gtk.MessageDialog(None, Gtk.DialogFlags.MODAL, Gtk.MessageType.INFO, Gtk.ButtonsType.OK, _('The following dependencies must be installed in order to use this application:') + '\n\n' + '\n'.join(arrDependenciesExternal) + '\n\n' +  _('The program will now exit.'))
                dlg.run()
                dlg.destroy()

                sys.exit(1)

        if len(arrDependenciesLocal) == 0:
            self.on_transaction_finished(None, 'exit-success')

    def on_error(self, pError):

        if isinstance(pError, aptdaemon.errors.NotAuthorizedError):

            self.pTransaction.cancel()

        else:

            try:

                pError.code

            except AttributeError:

                pass

            else:

                if 'org.freedesktop.DBus.Error.NoReply' not in str(pError):

                    pAptErrorDialog = aptdaemon.gtk3widgets.AptErrorDialog(pError)
                    pAptErrorDialog.run()
                    pAptErrorDialog.hide()

                    self.pTransaction.cancel()

    def on_transaction_finished(self, pTransaction, strExitState):

        if strExitState != 'exit-success':
            sys.exit(1)

        if self.bShowSplashScreen:
            GObject.timeout_add(1000, self.fnName)
        else:
            self.fnName()

    def set_label(self, strText):

        self.label.set_label(strText)

        while Gtk.events_pending():
            Gtk.main_iteration()

class Application:

    splashscreen = None
    m_pStatusBarLabel = None

    def __init__(self, bMinimiseToTray, bShowSplashScreen):

        for pProc in psutil.process_iter():

            if pProc.name() == APPNAME and pProc.pid != os.getpid():
                sys.exit(1)

        self.bMinimiseToTray = bMinimiseToTray
        self.bShowSplashScreen = bShowSplashScreen

        if hasattr(self, 'preinit'):
            self.preinit()

        self.splashscreen = SplashScreen(self._init, bShowSplashScreen)
        Gtk.main()

    def _init(self):

        global g_pWindow

        if hasattr(self, 'init'):
            self.init()

        gtkBuilder = Gtk.Builder()
        gtkBuilder.set_translation_domain(APPNAME)
        gtkBuilder.add_from_file(get_data_path('/usr/share/' + APPNAME + '/' + APPNAME + '.glade'))
        gtkBuilder.connect_signals(self)

        for pObject in gtkBuilder.get_objects():

            try:
                self.__dict__[moduleVar(Gtk.Buildable.get_name(pObject))] = pObject
            except:
                pass

        if hasattr(self, 'm_pWindow'):

            self.m_pToolbar.get_style_context().add_class(Gtk.STYLE_CLASS_PRIMARY_TOOLBAR)
            self.m_pWindow.set_icon_from_file(get_data_path('/usr/share/icons/hicolor/scalable/apps/' + APPNAME + '.svg'))

            strTitle = APPTITLE

            if APPDEBUG:
                strTitle += ' - DEBUGGING MODE'

            self.m_pWindow.set_title(strTitle)
            g_pWindow = self.m_pWindow
            self.m_pStatusBarLabel = self.m_pStatusBar.get_message_area().get_children()[0]

            if '--hide' in sys.argv and self.bMinimiseToTray:
                self.m_pWindow.hide()
            else:
                self.m_pWindow.show_all()

        if hasattr(self, '_postinit'):
            self._postinit()

        if hasattr(self, 'postinit'):
            self.postinit()

        if self.bShowSplashScreen:
            self.splashscreen.destroy()

    def set_splashscreen_label(self, strLabel):
        self.splashscreen.set_label(strLabel)

    def OnWindowDestroy(self, widget, data = None):
        Gtk.main_quit()

    def OnAboutClicked(self, widget, data = None):

        self.m_pAboutdialog = Gtk.AboutDialog()
        self.m_pAboutdialog.set_transient_for(g_pWindow)
        self.m_pAboutdialog.set_title(_('About ') + APPTITLE)
        self.m_pAboutdialog.set_license_type(Gtk.License.GPL_3_0)
        self.m_pAboutdialog.set_program_name(APPTITLE)
        self.m_pAboutdialog.set_copyright(APPAUTHOR + ' ' + (APPYEAR if APPYEAR[-2:] == APPVERSION[:2] else APPYEAR + '-20' + APPVERSION[:2]))
        self.m_pAboutdialog.set_comments(_(APPDESCRIPTION))
        self.m_pAboutdialog.set_authors([APPAUTHOR + ' &lt;' + APPMAIL + '&gt;'])
        self.m_pAboutdialog.set_translator_credits(_('translator-credits'))
        self.m_pAboutdialog.set_version(APPVERSION)
        self.m_pAboutdialog.set_website(APPURL)
        self.m_pAboutdialog.set_website_label(APPURL)
        self.m_pAboutdialog.set_logo(GdkPixbuf.Pixbuf().new_from_file(get_data_path('/usr/share/icons/hicolor/scalable/apps/' + APPNAME + '.svg')))
        self.m_pAboutdialog.run()
        self.m_pAboutdialog.destroy()

    def onWindowDeleteEvent(self, widget, event, data = None):
        return False

class Indicator(Application):

    def __init__(self):

        gi.require_version('AppIndicator3', '0.1')

        from gi.repository import AppIndicator3

        self.m_pStatus = AppIndicator3.IndicatorStatus

        if APPDEBUG:
            self.m_pIndicator = AppIndicator3.Indicator.new_with_path(APPNAME, APPNAME + '-active', AppIndicator3.IndicatorCategory.APPLICATION_STATUS, get_data_path('/usr/share/icons/ubuntu-mono-dark/status/24'))
        else:
            self.m_pIndicator = AppIndicator3.Indicator.new(APPNAME, APPNAME + '-active', AppIndicator3.IndicatorCategory.APPLICATION_STATUS)

        Application.__init__(self, False, False)

    def _postinit(self):

        self.m_pIndicator.set_attention_icon(APPNAME + '-attention')
        self.m_pIndicator.set_status(self.m_pStatus.ACTIVE)
        self.m_pIndicator.set_menu(self.m_pMenu)

class Dialog():

    nResponseID = Gtk.ResponseType.CANCEL

    def __init__(self, oParent, oData):

        global g_pWindow

        if hasattr(self, 'preinit'):
            self.preinit(oData)

        strFile = get_data_path('/usr/share/' + APPNAME + '/' + self.__class__.__module__.split('.')[-1].split('/')[-1] + '.glade')

        self.m_gtkBuilder = Gtk.Builder()
        self.m_gtkBuilder.set_translation_domain(APPNAME)
        self.m_gtkBuilder.add_from_file(strFile)
        self.m_gtkBuilder.connect_signals(self)

        for pObject in self.m_gtkBuilder.get_objects():

            if isinstance(pObject, Gtk.Buildable):

                self.__dict__[moduleVar(Gtk.Buildable.get_name(pObject))] = pObject

                if isinstance(pObject, Gtk.Notebook):

                    def getPageSensitive(pWidget, nPage):
                         return pWidget.get_tab_label(pWidget.get_nth_page(nPage)).get_sensitive()

                    def setPageSensitive(pWidget, nPage, bSensitive):
                         return pWidget.get_tab_label(pWidget.get_nth_page(nPage)).set_sensitive(bSensitive)

                    def onNotebookPageSelected(pWidget, pEvent):

                        nMouseX, nMouseY = pEvent.get_coords()

                        for nPage in range(0, pWidget.get_n_pages()):

                            pLabel = pWidget.get_tab_label(pWidget.get_nth_page(nPage))
                            nX, nY = pLabel.translate_coordinates(pWidget, 0, 0)
                            rcSize = pLabel.get_allocation()

                            if nMouseX >= nX and nMouseY >= nY and nMouseX <= nX + rcSize.width and nMouseY <= nY + rcSize.height and pWidget.getPageSensitive(nPage):
                                return False

                        return True

                    pObject.connect('button-press-event', onNotebookPageSelected)
                    pObject.getPageSensitive = types.MethodType(getPageSensitive, pObject)
                    pObject.setPageSensitive = types.MethodType(setPageSensitive, pObject)

        self.m_pDialog.set_transient_for(oParent if oParent else g_pWindow)
        self.m_pDialog.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.m_pDialog.set_skip_pager_hint(True)
        self.m_pDialog.set_skip_taskbar_hint(True)
        self.m_pDialog.set_destroy_with_parent(True)

        # FileChooserButtonEx
        with open(strFile) as oFile:

            oSoup = BeautifulSoup(oFile, 'xml')

            for oTag in oSoup.find_all('object', class_='GtkFileChooserButton', id=re.compile('^fileChooserButtonEx')):

                oFileChooserButton = self.__dict__[moduleVar(oTag['id'])]
                oParent = oFileChooserButton.get_parent()
                oFileChooserButton.destroy()

                oFileChooserButton = FileChooserButton()
                oParent.add(oFileChooserButton)

                for oProperty in oTag.parent.find_all('property'):

                    if oProperty['name'] == 'filter':
                        oFileChooserButton.set_property('filter', self.__dict__[moduleVar(oProperty.string)])
                    elif oProperty['name'] in ['left_attach', 'top_attach', 'width']:
                        oParent.child_set_property(oFileChooserButton, oProperty['name'].replace('_', '-'), int(oProperty.string))
                    else:
                        oFileChooserButton.set_property(oProperty['name'], oProperty.string)

                self.__dict__[moduleVar(oTag['id'])] = oFileChooserButton

        #~FileChooserButtonEx

        if hasattr(self, 'postinit'):
            self.postinit(oData)

        if self.m_pDialog.run() != -1:
            self.m_pDialog.destroy()

class Window():

    def __init__(self, data = None):

        global g_pWindow

        if hasattr(self, 'preinit'):
            self.preinit(data)

        self.m_gtkBuilder = Gtk.Builder()
        self.m_gtkBuilder.set_translation_domain(APPNAME)
        self.m_gtkBuilder.add_from_file(get_data_path('/usr/share/' + APPNAME + '/' + self.__class__.__module__.split('.')[-1].split('/')[-1] + '.glade'))
        self.m_gtkBuilder.connect_signals(self)

        for pObject in self.m_gtkBuilder.get_objects():

            if isinstance(pObject, Gtk.Buildable):

                self.__dict__[moduleVar(Gtk.Buildable.get_name(pObject))] = pObject

                if isinstance(pObject, Gtk.Notebook):

                    def getPageSensitive(pWidget, nPage):
                         return pWidget.get_tab_label(pWidget.get_nth_page(nPage)).get_sensitive()

                    def setPageSensitive(pWidget, nPage, bSensitive):
                         return pWidget.get_tab_label(pWidget.get_nth_page(nPage)).set_sensitive(bSensitive)

                    def onNotebookPageSelected(pWidget, pEvent):

                        nMouseX, nMouseY = pEvent.get_coords()

                        for nPage in range(0, pWidget.get_n_pages()):

                            pLabel = pWidget.get_tab_label(pWidget.get_nth_page(nPage))
                            nX, nY = pLabel.translate_coordinates(pWidget, 0, 0)
                            rcSize = pLabel.get_allocation()

                            if nMouseX >= nX and nMouseY >= nY and nMouseX <= nX + rcSize.width and nMouseY <= nY + rcSize.height and pWidget.getPageSensitive(nPage):
                                return False

                        return True

                    pObject.connect('button-press-event', onNotebookPageSelected)
                    pObject.getPageSensitive = types.MethodType(getPageSensitive, pObject)
                    pObject.setPageSensitive = types.MethodType(setPageSensitive, pObject)

                elif isinstance(pObject, Gtk.Toolbar):

                    self.m_pToolbar.get_style_context().add_class(Gtk.STYLE_CLASS_PRIMARY_TOOLBAR)

        g_pWindow = self.m_pWindow

        self.m_pWindow.set_position(Gtk.WindowPosition.CENTER)
        self.m_pWindow.set_skip_pager_hint(False)
        self.m_pWindow.set_skip_taskbar_hint(False)
        self.m_pWindow.set_type_hint(Gdk.WindowTypeHint.NORMAL)
        self.m_pWindow.set_icon_from_file(get_data_path('/usr/share/icons/hicolor/scalable/apps/' + APPNAME + '.svg'))
        self.m_pWindow.show_all()

        if hasattr(self, 'postinit'):
            self.postinit(data)

    def OnAboutClicked(self, widget, data = None):

        self.m_pAboutdialog = Gtk.AboutDialog()
        self.m_pAboutdialog.set_transient_for(g_pWindow)
        self.m_pAboutdialog.set_title(_('About ') + APPTITLE)
        self.m_pAboutdialog.set_license_type(Gtk.License.GPL_3_0)
        self.m_pAboutdialog.set_program_name(APPTITLE)
        self.m_pAboutdialog.set_copyright(APPAUTHOR + ' ' + (APPYEAR if APPYEAR[-2:] == APPVERSION[:2] else APPYEAR + '-20' + APPVERSION[:2]))
        self.m_pAboutdialog.set_comments(_(APPDESCRIPTION))
        self.m_pAboutdialog.set_authors([APPAUTHOR + ' &lt;' + APPMAIL + '&gt;'])
        self.m_pAboutdialog.set_translator_credits(_('translator-credits'))
        self.m_pAboutdialog.set_version(APPVERSION)
        self.m_pAboutdialog.set_website(APPURL)
        self.m_pAboutdialog.set_website_label(APPURL)
        self.m_pAboutdialog.set_logo(GdkPixbuf.Pixbuf().new_from_file(get_data_path('/usr/share/icons/hicolor/scalable/apps/' + APPNAME + '.svg')))
        self.m_pAboutdialog.run()
        self.m_pAboutdialog.destroy()

class FileChooserButton(Gtk.ButtonBox):

    oLabel = Gtk.Label('', xalign=0, margin_left=5)
    filter = GObject.Property(type=Gtk.FileFilter)
    title = GObject.Property(type=str)
    filename = GObject.Property(type=str)

    def __init__(self):

        Gtk.ButtonBox.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, layout_style=Gtk.ButtonBoxStyle.EXPAND)

        oButtonClear = Gtk.Button.new_from_icon_name('gtk-clear', Gtk.IconSize.BUTTON)
        oButtonClear.connect('clicked', self.onClear)
        oButtonOpen = Gtk.Button()
        oButtonOpen.connect('clicked', self.onOpen)
        oImage = Gtk.Image.new_from_icon_name('gtk-open', Gtk.IconSize.BUTTON)
        oBox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        oBox.pack_start(oImage, False, False, 0)
        oBox.pack_start(self.oLabel, True, True, 0)
        oButtonOpen.add(oBox)
        self.pack_start(oButtonOpen, True, True, 0)
        self.pack_start(oButtonClear, False, False, 0)
        self.set_homogeneous(False)
        self.show_all()
        self.connect('notify::filename', self.onNotifyFilename)

    def onClear(self, oWidget):

        self.filename = ''
        self.oLabel.set_text('')

    def onOpen(self, oWidget):

        oDlg = Gtk.FileChooserDialog(None, g_pWindow, Gtk.FileChooserAction.OPEN, (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OPEN, Gtk.ResponseType.APPLY))
        oDlg.add_filter(self.filter)
        oDlg.set_title(self.title)

        if self.filename:
            oDlg.set_filename(self.filename)

        nResponse = oDlg.run()

        if nResponse == Gtk.ResponseType.APPLY:

            self.filename = oDlg.get_filename()
            self.oLabel.set_text(os.path.basename(self.filename))

        oDlg.destroy()

    def onNotifyFilename(self, oObject, oGParamString):

        self.oLabel.set_text(os.path.basename(self.filename))


